#!/usr/bin/python
# -*- coding: utf-8 -*-

from timeit import default_timer as timer

import pandas as pd
import numpy as np

import re
from bs4 import BeautifulSoup
from nltk.stem import SnowballStemmer
import nltk

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer

from sklearn import cross_validation
from sklearn import svm

### PRE PROCESSAMENTO E LIMPEZA ########################################################################################

stemmer = SnowballStemmer("english")

# Transforma um review em uma string limpa de palavras
def review_to_features(raw):
    # 1. Limpa tags HTML
    review_text = BeautifulSoup(raw, "html.parser").get_text()
    #
    # 2. Remove caracteres inválidos (review em inglês, então não possuí acento)
    valid_characters = re.sub("[^a-zA-Z0-9']", " ", review_text)
    #
    # 3. Etiquetando pela função gramatical
    tags = nltk.pos_tag(valid_characters.split())
    #
    # 4. Filtrando só os adjetivos e adverbios, deixando em letras minusculas
    # nltk.help.upenn_tagset()
    adjectives = [w[0].lower() for w in tags if w[1] in ['JJ', 'JJR', 'JJS', 'RB', 'RBR', 'RBS']]
    #
    # 5. Aplicando o SnowballStemmer
    tokens = [stemmer.stem(w) for w in adjectives]
    # 6. Juntando em uma string separada por espaços e retornando
    return (" ".join(tokens))


# Carregando o arquivo em um pandas.core.frames.DataFrame
print("Loading the reviews training set")
train = pd.read_csv("./data/labeledTrainData.tsv", header = 0, delimiter = "\t", quoting = 3)
#train = pd.read_csv("./data/sampleLabeled.tsv", header = 0, delimiter = "\t", quoting = 3)

# Limpando os reviews com a função
print("Cleaning and parsing the reviews, it may take a while...")
start = timer()
clean_train = train["review"].apply(review_to_features)
print("Train data cleaned and parsed, elapsed time: %.2f " % (timer() - start) + "seconds\n")

# Como demora muito para fazer essa limpeza, foi exportado o dataframe, carregando a versão extraida
train = pd.read_csv("labeled_adjectives.tsv", header = 0, delimiter = "\t")
clean_train = train["adjectives"].astype("unicode")

### SVM LINEAR & POLY 2 COM BAG OF WORDS ###############################################################################

# Construindo uma bag of words com CountVectorizer e outra com TfidfVectorizer
countBag = CountVectorizer(analyzer = "word", max_features = 2000).fit_transform(clean_train).toarray()
tfidfBag = TfidfVectorizer(analyzer = "word", max_features = 2000).fit_transform(clean_train).toarray()

# Fazendo validação cruzada com 40% para teste, 60% para treino
print("SVM com Bag of words\n")

# 1. Linear SVM - Count Bag of words
print("Linear SVM - Count Bag of words:")
start = timer()
xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(countBag, train["sentiment"], test_size = 0.4)
svm_linear_count_bag = svm.SVC(kernel = "linear").fit(xtrain, ytrain)
print("    score-in    %.4f" % svm_linear_count_bag.score(xtrain, ytrain))
print("    score-out   %.4f" % svm_linear_count_bag.score(xtest, ytest))
print("    total time  %.2f " % (timer() - start) + "seconds\n")

# 2. Linear SVM - Tf-idf Bag of words
print("Linear SVM - Tf-idf Bag of words:")
start = timer()
xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(tfidfBag, train["sentiment"], test_size = 0.4)
svm_linear_tfidf_bag= svm.SVC(kernel = "linear").fit(xtrain, ytrain)
print("    score-in    %.4f" % svm_linear_tfidf_bag.score(xtrain, ytrain))
print("    score-out   %.4f" % svm_linear_tfidf_bag.score(xtest, ytest))
print("    total time  %.2f " % (timer() - start) + "seconds\n")
#
# 3. Poly2 SVM - Count Bag of words
#print("Polynomial 2 SVM - Count Bag of words")
#start = timer()
#xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(countBag, train["sentiment"], test_size = 0.4)
#svm_poly2_count_bag = svm.SVC(kernel = "poly", degree = 2).fit(xtrain, ytrain)
#print("    score-in    %.4f" % svm_poly2_count_bag.score(xtrain, ytrain))
#print("    score-out   %.4f" % svm_poly2_count_bag.score(xtest, ytest))
#print("    total time  %.2f " % (timer() - start) + "seconds\n")
#
# 4. Poly2 SVM - Tf-idf Bag of words
#print("Polynomial 2 SVM - Tf-idf Bag of words")
#start = timer()
#xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(tfidfBag, train["sentiment"], test_size = 0.4)
#svm_poly2_tfidf_bag= svm.SVC(kernel = "poly", degree = 2).fit(xtrain, ytrain)
#print("    score-in    %.4f" % svm_poly2_tfidf_bag.score(xtrain, ytrain))
#print("    score-out   %.4f" % svm_poly2_tfidf_bag.score(xtest, ytest))
#print("    total time  %.2f " % (timer() - start) + "seconds\n")

### SVM COM STRING KERNELS #############################################################################################

def gap_weighted_subsequence(lamb, p):
    # Memoização com dicionário
    cache = {}
    #
    # Cálcula a distância entra a string s1 e s2
    def gws(s1, s2, lamb, p):
        key = (s1 ,s2) if s1 > s2 else (s2, s1)
        if not key in cache:
            # dps: matriz s1 x s2 com lamb^2 se igual e 0 caso contrário
            dps = []
            for i in xrange(len(s1)):
                dps.append([lamb**2 if s1[i] == s2[j] else 0 for j in xrange(len(s2))])
            #dp: matriz de 0 de tamanho len(s1)+1 x len(s2)+1
            dp = []
            for i in xrange(len(s1) + 1):
                dp.append([0] * (len(s2) + 1))
            #k: array de 0 de tamanho p+1
            k = [0] * (p + 1)
            for l in xrange(2, p + 1):
                for i in xrange(len(s1)):
                    for j in xrange(len(s2)):
                        dp[i+1][j+1] = dps[i][j] + lamb * dp[i][j+1] + lamb * dp[i+1][j] - lamb**2 * dp[i][j]
                        if s1[i] == s2[j]:
                            dps[i][j] = lamb**2 * dp[i][j]
                            k[l] = k[l] + dps[i][j]
            cache[key] = k[p]
        return cache[key]
    return lambda s1, s2: gws(s1,s2,lamb,p)/(gws(s1,s1,lamb,p) * gws(s2,s2,lamb,p))**0.5

def split(string):
    return string.split(" ")

def dump(dps, dp, k):
    print("DPS")
    for row in dps:
       print row
    print("DP")
    for row in dp:
       print row
    print("K")
    print k
    print("\n")

def gap_weighted_word(lamb, p):
    def gww_kernel(s1, s2, lamb, p):
        # dps: matriz s1 x s2 com lamb^2 se igual e 0 caso contrário
        dps = []
        for i in xrange(len(s1)):
            dps.append([lamb**2 if s1[i] == s2[j] else 0 for j in xrange(len(s2))])
        #dp: matriz de 0 de tamanho len(s1)+1 x len(s2)+1
        dp = []
        for i in xrange(len(s1) + 1):
            dp.append([0] * (len(s2) + 1))
        #k: array de 0 de tamanho p+1
        k = [0] * (p + 1)
        if s1 != s2:
           dump(dps, dp, k)
        for l in xrange(1, p + 1):
            for i in xrange(len(s1)):
                for j in xrange(len(s2)):
                    dp[i+1][j+1] = dps[i][j] + lamb * dp[i][j+1] + lamb * dp[i+1][j] - lamb**2 * dp[i][j]
                    if s1[i] == s2[j]:
                        dps[i][j] = lamb**2 * dp[i][j]
                        k[l] = k[l] + dps[i][j]
                        if s1 != s2:
                            dump(dps, dp, k)
        return k[p]
    def normalize(s1, s2):
        s1s2 = gww_kernel(s1,s2,lamb,p)
        s1s1 = gww_kernel(s1,s1,lamb,p)
        s2s2 = gww_kernel(s2,s2,lamb,p)
        return s1s2/(s1s1 * s2s2)**0.5 if s1s1 * s2s2 != 0 else 0
    return lambda s1, s2: normalize(s1,s2)

# Cálcula a gram matrix para o conjunto de documentos x1 e x2
def gram_matrix(algorithm, x1, x2):
    gram_matrix = np.zeros((len(x1), len(x2)), dtype = np.float32)
    for i, x1row in enumerate(x1):
        for j, x2row in enumerate(x2):
            if algorithm == "gws":
                gram_matrix[i][j] = gws(x1row, x2row)
            elif algorithm == "gww":
                gram_matrix[i][j] = gww(x1row, x2row)
    return gram_matrix

# Fazendo validação cruzada com 40% para teste
print("SVM com String Kernels\n")

## 1. SVM - Gap weighted subsequence
#print("Gap weighted subsequence kernel:")
#start = timer()
#gws = gap_weighted_subsequence(0.5, 3)
#xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(clean_train, train["sentiment"], test_size = 0.4)
#gws_train_gm = gram_matrix("gws", xtrain, xtrain)
#svm_gws = svm.SVC(kernel = "precomputed").fit(gws_train_gm, ytrain)
#print("    score-in    %.4f" % svm_gws.score(gws_train_gm, ytrain))
#gws_test_gm = gram_matrix("gws", xtest, xtrain)
#print("    score-out   %.4f" % svm_gws.score(gws_test_gm, ytest))
#print("    total time  %.2f " % (timer() - start) + "seconds\n")
#
## 2. SVM - Gap weighted words (Gubi's Kernel)
#print("Gap weighted words - Gubi's kernel:")
start = timer()
gww = gap_weighted_word(0.5, 3)
xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(clean_train[0:5000].apply(split), train["sentiment"][0:5000], test_size = 0.4)
gww_train_gm = gram_matrix("gww", xtrain, xtrain)
svm_gww = svm.SVC(kernel = "precomputed").fit(gww_train_gm, ytrain)
print("    score-in    %.4f" % svm_gww.score(gww_train_gm, ytrain))
gww_test_gm = gram_matrix("gww", xtest, xtrain)
print("    score-out   %.4f" % svm_gww.score(gww_test_gm, ytest))
print("    total time  %.2f " % (timer() - start) + "seconds\n")

# 2. SVM - Gap weighted words (Gubi's Kernel)
print("Gap weighted words - Gubi's kernel:")
for i in xrange(2, 6):
   start = timer()
   print("lambda = 0.5, e p = %d" % i)
   gww = gap_weighted_word(0.5, i)
   xtrain, xtest, ytrain, ytest = cross_validation.train_test_split(clean_train[0:5000].apply(split), train["sentiment"][0:5000], test_size = 0.4)
   gww_train_gm = gram_matrix("gww", xtrain, xtrain)
   svm_gww = svm.SVC(kernel = "precomputed").fit(gww_train_gm, ytrain)
   print("    score-in    %.4f" % svm_gww.score(gww_train_gm, ytrain))
   gww_test_gm = gram_matrix("gww", xtest, xtrain)
   print("    score-out   %.4f" % svm_gww.score(gww_test_gm, ytest))
   print("    total time  %.2f " % (timer() - start) + "seconds\n")

